/* Copyright (c) 2000-2010 Wolfgang Hoermann and Josef Leydold */
/* Department of Statistics and Mathematics, WU Wien, Austria  */

#include <unur_source.h>
#include <distr/distr.h>
#include <distr/distr_source.h>
#include <distr/cont.h>
#include <urng/urng.h>
#include "unur_methods_source.h"
#include "x_gen_source.h"
#include "ssr.h"
#include "ssr_struct.h"
#ifdef UNUR_ENABLE_INFO
#  include <tests/unuran_tests.h>
#endif
#define SSR_VARFLAG_VERIFY   0x002u    
#define SSR_VARFLAG_SQUEEZE  0x004u    
#define SSR_VARFLAG_MIRROR   0x008u    
#define SSR_DEBUG_REINIT    0x00000010u   
#define SSR_SET_CDFMODE      0x001u    
#define SSR_SET_PDFMODE      0x002u    
#define GENTYPE "SSR"          
static struct unur_gen *_unur_ssr_init( struct unur_par *par );
static int _unur_ssr_reinit( struct unur_gen *gen );
static struct unur_gen *_unur_ssr_create( struct unur_par *par );
static int _unur_ssr_check_par( struct unur_gen *gen );
static struct unur_gen *_unur_ssr_clone( const struct unur_gen *gen );
static void _unur_ssr_free( struct unur_gen *gen);
static double _unur_ssr_sample( struct unur_gen *gen );
static double _unur_ssr_sample_check( struct unur_gen *gen );
static int _unur_ssr_hat( struct unur_gen *gen );
#ifdef UNUR_ENABLE_LOGGING
static void _unur_ssr_debug_init( const struct unur_gen *gen, int is_reinit );
#endif
#ifdef UNUR_ENABLE_INFO
static void _unur_ssr_info( struct unur_gen *gen, int help );
#endif
#define DISTR_IN  distr->data.cont      
#define PAR       ((struct unur_ssr_par*)par->datap) 
#define GEN       ((struct unur_ssr_gen*)gen->datap) 
#define DISTR     gen->distr->data.cont 
#define BD_LEFT   domain[0]             
#define BD_RIGHT  domain[1]             
#define SAMPLE    gen->sample.cont           
#define PDF(x)    _unur_cont_PDF((x),(gen->distr))    
#define _unur_ssr_getSAMPLE(gen) \
   ( ((gen)->variant & SSR_VARFLAG_VERIFY) \
     ? _unur_ssr_sample_check : _unur_ssr_sample )
#define SQRT2    1.4142135623731
struct unur_par *
unur_ssr_new( const struct unur_distr *distr )
{ 
  struct unur_par *par;
  _unur_check_NULL( GENTYPE,distr,NULL );
  if (distr->type != UNUR_DISTR_CONT) {
    _unur_error(GENTYPE,UNUR_ERR_DISTR_INVALID,""); return NULL; }
  COOKIE_CHECK(distr,CK_DISTR_CONT,NULL);
  if (DISTR_IN.pdf == NULL) {
    _unur_error(GENTYPE,UNUR_ERR_DISTR_REQUIRED,"PDF"); 
    return NULL;
  }
  par = _unur_par_new( sizeof(struct unur_ssr_par) );
  COOKIE_SET(par,CK_SSR_PAR);
  par->distr    = distr;            
  PAR->Fmode     = -1.;             
  PAR->fm        = -1.;             
  PAR->um        = -1.;             
  par->method   = UNUR_METH_SSR;    
  par->variant  = 0u;               
  par->set      = 0u;                   
  par->urng     = unur_get_default_urng(); 
  par->urng_aux = NULL;                    
  par->debug    = _unur_default_debugflag; 
  par->init = _unur_ssr_init;
  return par;
} 
int 
unur_ssr_set_cdfatmode( struct unur_par *par, double Fmode )
{
  _unur_check_NULL( GENTYPE, par, UNUR_ERR_NULL );
  _unur_check_par_object( par, SSR );
  if (Fmode < 0. || Fmode > 1.) {
    _unur_warning(GENTYPE,UNUR_ERR_PAR_SET,"CDF(mode)");
    return UNUR_ERR_PAR_SET;
  }
  PAR->Fmode = Fmode;
  par->set |= SSR_SET_CDFMODE;
  return UNUR_SUCCESS;
} 
int 
unur_ssr_set_pdfatmode( UNUR_PAR *par, double fmode )
{
  _unur_check_NULL( GENTYPE, par, UNUR_ERR_NULL );
  _unur_check_par_object( par, SSR );
  if (fmode <= 0.) {
    _unur_warning(GENTYPE,UNUR_ERR_PAR_SET,"PDF(mode)");
    return UNUR_ERR_PAR_SET;
  }
  if (!_unur_isfinite(fmode)) {
    _unur_warning(GENTYPE,UNUR_ERR_PAR_SET,"PDF(mode) overflow");
    return UNUR_ERR_PAR_SET;
  }
  PAR->fm = fmode;
  PAR->um = sqrt(fmode);
  par->set |= SSR_SET_PDFMODE;
  return UNUR_SUCCESS;
} 
int
unur_ssr_set_verify( struct unur_par *par, int verify )
{
  _unur_check_NULL( GENTYPE, par, UNUR_ERR_NULL );
  _unur_check_par_object( par, SSR );
  par->variant = (verify) ? (par->variant | SSR_VARFLAG_VERIFY) : (par->variant & (~SSR_VARFLAG_VERIFY));
  return UNUR_SUCCESS;
} 
int
unur_ssr_chg_verify( struct unur_gen *gen, int verify )
{
  _unur_check_NULL( GENTYPE, gen, UNUR_ERR_NULL );
  _unur_check_gen_object( gen, SSR, UNUR_ERR_GEN_INVALID );
  if (SAMPLE == _unur_sample_cont_error) 
    return UNUR_FAILURE;
  if (verify)
    gen->variant |= SSR_VARFLAG_VERIFY;
  else
    gen->variant &= ~SSR_VARFLAG_VERIFY;
  SAMPLE = _unur_ssr_getSAMPLE(gen);
  return UNUR_SUCCESS;
} 
int 
unur_ssr_set_usesqueeze( struct unur_par *par, int usesqueeze )
{
  _unur_check_NULL( GENTYPE, par, UNUR_ERR_NULL );
  _unur_check_par_object( par, SSR );
  par->variant = (usesqueeze) 
    ? (par->variant | SSR_VARFLAG_SQUEEZE) 
    : (par->variant & (~SSR_VARFLAG_SQUEEZE));
  return UNUR_SUCCESS;
} 
int
unur_ssr_chg_cdfatmode( struct unur_gen *gen, double Fmode )
{
  _unur_check_NULL( GENTYPE, gen, UNUR_ERR_NULL );
  _unur_check_gen_object( gen, SSR, UNUR_ERR_GEN_INVALID );
  if (Fmode < 0. || Fmode > 1.) {
    _unur_warning(gen->genid,UNUR_ERR_PAR_SET,"CDF(mode)");
    return UNUR_ERR_PAR_SET;
  }
  GEN->Fmode = Fmode;
  gen->set |= SSR_SET_CDFMODE;
  return UNUR_SUCCESS;
} 
int
unur_ssr_chg_pdfatmode( struct unur_gen *gen, double fmode )
{
  _unur_check_NULL( GENTYPE, gen, UNUR_ERR_NULL );
  _unur_check_gen_object( gen, SSR, UNUR_ERR_GEN_INVALID );
  if (fmode <= 0.) {
    _unur_warning(gen->genid,UNUR_ERR_PAR_SET,"PDF(mode)");
    return UNUR_ERR_PAR_SET;
  }
  if (!_unur_isfinite(fmode)) {
    _unur_warning(GENTYPE,UNUR_ERR_PAR_SET,"PDF(mode) overflow");
    return UNUR_ERR_PAR_SET;
  }
  GEN->fm = fmode;
  GEN->um = sqrt(fmode);
  gen->set |= SSR_SET_PDFMODE;
  return UNUR_SUCCESS;
} 
struct unur_gen *
_unur_ssr_init( struct unur_par *par )
{ 
  struct unur_gen *gen;
  CHECK_NULL(par,NULL);
  if ( par->method != UNUR_METH_SSR ) {
    _unur_error(GENTYPE,UNUR_ERR_PAR_INVALID,"");
    return NULL; }
  COOKIE_CHECK(par,CK_SSR_PAR,NULL);
  if (! (par->set & SSR_SET_CDFMODE))
    par->variant &= ~SSR_VARFLAG_SQUEEZE;
  gen = _unur_ssr_create(par);
  _unur_par_free(par);
  if (!gen) return NULL;
  if (_unur_ssr_check_par(gen) != UNUR_SUCCESS) {
    _unur_ssr_free(gen); return NULL;
  }
  if (_unur_ssr_hat(gen)!=UNUR_SUCCESS) {
    _unur_ssr_free(gen); return NULL;
  }
#ifdef UNUR_ENABLE_LOGGING
    if (gen->debug) _unur_ssr_debug_init(gen, FALSE);
#endif
  return gen;
} 
int
_unur_ssr_reinit( struct unur_gen *gen )
{
  int rcode;
  if ( (rcode = _unur_ssr_check_par(gen)) != UNUR_SUCCESS)
    return rcode;
  rcode = _unur_ssr_hat( gen );
  SAMPLE = _unur_ssr_getSAMPLE(gen);
#ifdef UNUR_ENABLE_LOGGING
  if (gen->debug & SSR_DEBUG_REINIT) _unur_ssr_debug_init(gen,TRUE);
#endif
  return rcode;
} 
struct unur_gen *
_unur_ssr_create( struct unur_par *par )
{
  struct unur_gen *gen;
  CHECK_NULL(par,NULL);  COOKIE_CHECK(par,CK_SSR_PAR,NULL);
  gen = _unur_generic_create( par, sizeof(struct unur_ssr_gen) );
  COOKIE_SET(gen,CK_SSR_GEN);
  gen->genid = _unur_set_genid(GENTYPE);
  SAMPLE = _unur_ssr_getSAMPLE(gen);
  gen->destroy = _unur_ssr_free;
  gen->clone = _unur_ssr_clone;
  gen->reinit = _unur_ssr_reinit;
  GEN->Fmode = PAR->Fmode;            
  GEN->fm    = PAR->fm;               
  GEN->um    = PAR->um;               
#ifdef UNUR_ENABLE_INFO
  gen->info = _unur_ssr_info;
#endif
  return gen;
} 
int
_unur_ssr_check_par( struct unur_gen *gen )
{
  if (!(gen->distr->set & UNUR_DISTR_SET_MODE)) {
    _unur_warning(GENTYPE,UNUR_ERR_DISTR_REQUIRED,"mode: try finding it (numerically)");
    if (unur_distr_cont_upd_mode(gen->distr)!=UNUR_SUCCESS) {
      _unur_error(GENTYPE,UNUR_ERR_DISTR_REQUIRED,"mode");
      return UNUR_ERR_DISTR_REQUIRED;
    }
  }
  if (!(gen->distr->set & UNUR_DISTR_SET_PDFAREA)) {
    if (unur_distr_cont_upd_pdfarea(gen->distr)!=UNUR_SUCCESS) {
      _unur_error(GENTYPE,UNUR_ERR_DISTR_REQUIRED,"area below PDF");
      return UNUR_ERR_DISTR_REQUIRED;
    }
  }
  if ( (DISTR.mode < DISTR.BD_LEFT) ||
       (DISTR.mode > DISTR.BD_RIGHT) ) {
    _unur_warning(GENTYPE,UNUR_ERR_GEN_DATA,"area and/or CDF at mode");
    DISTR.mode = _unur_max(DISTR.mode,DISTR.BD_LEFT);
    DISTR.mode = _unur_min(DISTR.mode,DISTR.BD_RIGHT);
  }
  return UNUR_SUCCESS;
} 
struct unur_gen *
_unur_ssr_clone( const struct unur_gen *gen )
{ 
#define CLONE  ((struct unur_ssr_gen*)clone->datap)
  struct unur_gen *clone;
  CHECK_NULL(gen,NULL);  COOKIE_CHECK(gen,CK_SSR_GEN,NULL);
  clone = _unur_generic_clone( gen, GENTYPE );
  return clone;
#undef CLONE
} 
void
_unur_ssr_free( struct unur_gen *gen )
{ 
  if( !gen ) 
    return;
  if ( gen->method != UNUR_METH_SSR ) {
    _unur_warning(gen->genid,UNUR_ERR_GEN_INVALID,"");
    return; }
  COOKIE_CHECK(gen,CK_SSR_GEN,RETURN_VOID);
  SAMPLE = NULL;   
  _unur_generic_free(gen);
} 
double
_unur_ssr_sample( struct unur_gen *gen )
{ 
  double U,V,X,xx,y;
  CHECK_NULL(gen,INFINITY);  COOKIE_CHECK(gen,CK_SSR_GEN,INFINITY);
  while (1) {
    while ( _unur_iszero(U = GEN->Aleft + _unur_call_urng(gen->urng) * GEN->Ain) );
    if (U < GEN->al) {        
      X = - GEN->vl * GEN->vl / U;
      y = (U / GEN->vl);
      y = y*y;
    }
    else if (U <= GEN->ar) {  
      X = GEN->xl + (U-GEN->al)/GEN->fm;
      y = GEN->fm;
    }
    else {                   
      X = GEN->vr * GEN->vr / (GEN->um * GEN->vr - (U-GEN->ar));
      y = (GEN->A - U) / GEN->vr;
      y = y*y;
    }
    V = _unur_call_urng(gen->urng);
    y *= V;
    if (gen->variant & SSR_VARFLAG_SQUEEZE) {
      xx = 2 * X;
      if ( xx >= GEN->xl && xx <= GEN->xr && y <= GEN->fm/4. )
	return (X + DISTR.mode);
    }
    X += DISTR.mode;
    if (y <= PDF(X))
      return X;
  }
} 
double
_unur_ssr_sample_check( struct unur_gen *gen )
{ 
  double U,V,X,xx,fx,y;
  CHECK_NULL(gen,INFINITY);  COOKIE_CHECK(gen,CK_SSR_GEN,INFINITY);
  while (1) {
    while ( _unur_iszero(U = GEN->Aleft + _unur_call_urng(gen->urng) * GEN->Ain) );
    if (U < GEN->al) {        
      X = - GEN->vl * GEN->vl / U;
      y = (U / GEN->vl);
      y = y*y;
    }
    else if (U <= GEN->ar) {  
      X = GEN->xl + (U-GEN->al)/GEN->fm;
      y = GEN->fm;
    }
    else {                   
      X = GEN->vr * GEN->vr / (GEN->um * GEN->vr - (U-GEN->ar));
      y = (GEN->A - U) / GEN->vr;
      y = y*y;
    }
    fx = PDF(X + DISTR.mode);
    if ( (1.+UNUR_EPSILON) * y < fx )
      _unur_error(gen->genid,UNUR_ERR_GEN_CONDITION,"PDF(x) > hat(x)");
    V = _unur_call_urng(gen->urng);
    y *= V;
    if (gen->variant & SSR_VARFLAG_SQUEEZE) {
      xx = 2 * X;
      if ( xx >= GEN->xl && xx <= GEN->xr ) {
	if ( fx < (1.-UNUR_EPSILON) * GEN->fm/4. )
	  _unur_error(gen->genid,UNUR_ERR_GEN_CONDITION,"PDF(x) < squeeze(x)");
	if ( y <= GEN->fm/4. )
	  return (X + DISTR.mode);
      }
    }
    X += DISTR.mode;
    if (y <= fx)
      return X;
  }
} 
int
_unur_ssr_hat( struct unur_gen *gen )
{ 
  double vm, fm;             
  double left,right;
  CHECK_NULL( gen, UNUR_ERR_NULL );
  COOKIE_CHECK( gen,CK_SSR_GEN, UNUR_ERR_COOKIE );
  if (!(gen->set & SSR_SET_PDFMODE)) {
    fm = PDF(DISTR.mode);
    if (fm <= 0.) {
      _unur_warning(gen->genid,UNUR_ERR_GEN_DATA,"PDF(mode) <= 0.");
      return UNUR_ERR_GEN_DATA;
    }
    if (!_unur_isfinite(fm)) {
      _unur_warning(GENTYPE,UNUR_ERR_PAR_SET,"PDF(mode) overflow");
      return UNUR_ERR_PAR_SET;
    }
    GEN->fm = fm;        
    GEN->um = sqrt(fm);  
  }
  vm = DISTR.area / GEN->um;
  if (gen->set & SSR_SET_CDFMODE) {
    GEN->vl = -GEN->Fmode * vm;
    GEN->vr = vm + GEN->vl;
    GEN->xl = GEN->vl/GEN->um;
    GEN->xr = GEN->vr/GEN->um;
    GEN->A  = 2 * DISTR.area;
    GEN->al = (DISTR.BD_LEFT  < DISTR.mode) ? (GEN->Fmode * DISTR.area) : 0.;
    GEN->ar = (DISTR.BD_RIGHT > DISTR.mode) ? (GEN->al + DISTR.area) : GEN->A;
    if ( (DISTR.BD_LEFT > -INFINITY) &&
	 (DISTR.BD_LEFT < DISTR.mode) )
      GEN->Aleft = GEN->vl * GEN->vl / (DISTR.mode - DISTR.BD_LEFT);
    else
      GEN->Aleft = 0.;
    if ( (DISTR.BD_RIGHT < INFINITY) &&
	 (DISTR.BD_RIGHT > DISTR.mode) )
      GEN->Ain = GEN->A - GEN->vr * GEN->vr / (DISTR.BD_RIGHT - DISTR.mode);
    else
      GEN->Ain = GEN->A;
    GEN->Ain -= GEN->Aleft;
  }
  else {
    GEN->vl = -vm;
    GEN->vr = vm;
    GEN->xl = GEN->vl/GEN->um;
    GEN->xr = GEN->vr/GEN->um;
    GEN->A  = 4 * DISTR.area;
    GEN->al = DISTR.area;
    GEN->ar = 3 * DISTR.area;
    if (DISTR.BD_LEFT > -INFINITY) {
      left = DISTR.BD_LEFT - DISTR.mode;
      GEN->Aleft = (GEN->xl > left) 
	? (GEN->vl * GEN->vl / (-left)) 
	: (GEN->al + GEN->fm * (left - GEN->xl));
    }
    else 
      GEN->Aleft = 0.;
    if (DISTR.BD_RIGHT < INFINITY) {
      right = DISTR.BD_RIGHT - DISTR.mode;
      GEN->Ain = (GEN->xr < right) 
	? (GEN->A - GEN->vr * GEN->vr / right)
	: (GEN->ar - GEN->fm * (GEN->xr - right));
    }
    else 
      GEN->Ain = GEN->A;
    GEN->Ain -= GEN->Aleft;
  }
#ifdef UNUR_ENABLE_LOGGING
#endif
  return UNUR_SUCCESS;
} 
#ifdef UNUR_ENABLE_LOGGING
static void
_unur_ssr_debug_init( const struct unur_gen *gen, int is_reinit )
{
  FILE *LOG;
  CHECK_NULL(gen,RETURN_VOID);  COOKIE_CHECK(gen,CK_SSR_GEN,RETURN_VOID);
  LOG = unur_get_stream();
  fprintf(LOG,"%s:\n",gen->genid);
  if (!is_reinit) {
    fprintf(LOG,"%s: type    = continuous univariate random variates\n",gen->genid);
    fprintf(LOG,"%s: method  = ssr (simple universal transformed density rection)\n",gen->genid);
  }
  else
    fprintf(LOG,"%s: reinit!\n",gen->genid);
  fprintf(LOG,"%s:\n",gen->genid);
  _unur_distr_cont_debug( gen->distr, gen->genid );
  fprintf(LOG,"%s: sampling routine = _unur_ssr_sample",gen->genid);
  if (gen->variant & SSR_VARFLAG_VERIFY)
    fprintf(LOG,"_check");
  fprintf(LOG,"()\n%s:\n",gen->genid);
  if (gen->set & SSR_SET_CDFMODE)
    fprintf(LOG,"%s: CDF at mode = %g\n",gen->genid,GEN->Fmode);
  else
    fprintf(LOG,"%s: CDF at mode unknown\n",gen->genid);
  if (gen->variant & SSR_VARFLAG_SQUEEZE)
    fprintf(LOG,"%s: use universal squeeze\n",gen->genid);
  else
    fprintf(LOG,"%s: no (universal) squeeze\n",gen->genid);
  fprintf(LOG,"%s:\n",gen->genid);
  fprintf(LOG,"%s: parts:\n",gen->genid);
  fprintf(LOG,"%s:\txl = %g\n",gen->genid,GEN->xl);
  fprintf(LOG,"%s:\txr = %g\n",gen->genid,GEN->xr);
  fprintf(LOG,"%s:\n",gen->genid);
  fprintf(LOG,"%s: PDF at mode:\n",gen->genid);
  fprintf(LOG,"%s:\tfm = %g\n",gen->genid,GEN->fm);
  fprintf(LOG,"%s:\tum = %g\n",gen->genid,GEN->um);
  fprintf(LOG,"%s:\n",gen->genid);
  fprintf(LOG,"%s: areas:\n",gen->genid);
  fprintf(LOG,"%s:\t    al = %g\n",gen->genid,GEN->al);
  fprintf(LOG,"%s:\t    ar = %g\n",gen->genid,GEN->ar);
  fprintf(LOG,"%s:\t Aleft = %g\n",gen->genid,GEN->Aleft);
  fprintf(LOG,"%s:\t   Ain = %g\n",gen->genid,GEN->Ain);
  fprintf(LOG,"%s:\tAtotal = %g\n",gen->genid,GEN->A);
  fprintf(LOG,"%s:\n",gen->genid);
} 
#endif   
#ifdef UNUR_ENABLE_INFO
void
_unur_ssr_info( struct unur_gen *gen, int help )
{
  struct unur_string *info = gen->infostr;
  struct unur_distr *distr = gen->distr;
  int samplesize = 10000;
  double rc, rc_approx;
  _unur_string_append(info,"generator ID: %s\n\n", gen->genid);
  _unur_string_append(info,"distribution:\n");
  _unur_distr_info_typename(gen);
  _unur_string_append(info,"   functions = PDF\n");
  _unur_string_append(info,"   domain    = (%g, %g)\n", DISTR.domain[0],DISTR.domain[1]);
  _unur_string_append(info,"   mode      = %g   %s\n", DISTR.mode,
		      (distr->set & UNUR_DISTR_SET_MODE_APPROX) ? "[numeric.]" : "");
  _unur_string_append(info,"   area(PDF) = %g\n", DISTR.area);
  if (gen->set & SSR_SET_CDFMODE)
    _unur_string_append(info,"   F(mode)   = %g\n", GEN->Fmode); 
  else
    _unur_string_append(info,"   F(mode)   = [unknown]\n"); 
  if (help) {
    if ( distr->set & UNUR_DISTR_SET_MODE_APPROX ) 
      _unur_string_append(info,"\n[ Hint: %s ]\n",
			  "You may provide the \"mode\"");
  }
  _unur_string_append(info,"\n");
  _unur_string_append(info,"method: SSR (Simple Ratio-Of-Uniforms)\n");
  if (gen->set & SSR_SET_CDFMODE)
    _unur_string_append(info,"   use CDF at mode\n");
  if (gen->variant & SSR_VARFLAG_SQUEEZE)
    _unur_string_append(info,"   use squeeze\n");
  _unur_string_append(info,"\n");
  _unur_string_append(info,"performance characteristics:\n");
  rc = (gen->set & SSR_SET_CDFMODE) ? 2. : 4.;
  if (_unur_isfinite(DISTR.BD_RIGHT) || _unur_isfinite(DISTR.BD_LEFT)) {
    rc_approx = unur_test_count_urn(gen,samplesize,0,NULL)/(2.*samplesize);
    _unur_string_append(info,"   rejection constant <= %g  [approx. = %.2f]\n", rc,rc_approx);
  }
  else {
    _unur_string_append(info,"   rejection constant = %g\n", rc);
  }
  _unur_string_append(info,"\n");
  if (help) {
    _unur_string_append(info,"parameters:\n");
    if (gen->set & SSR_SET_CDFMODE)
      _unur_string_append(info,"   cdfatmode = %g\n", GEN->Fmode); 
    else
      _unur_string_append(info,"   cdfatmode = [not set]\n"); 
    if (gen->variant & SSR_VARFLAG_SQUEEZE)
      _unur_string_append(info,"   usesqueeze\n");
    if (gen->variant & SSR_VARFLAG_VERIFY)
      _unur_string_append(info,"   verify = on\n");
    _unur_string_append(info,"\n");
  }
  if (help) {
    if ( !(gen->set & SSR_SET_CDFMODE))
      _unur_string_append(info,"[ Hint: %s ]\n",
			  "You can set \"cdfatmode\" to reduce the rejection constant.");
    _unur_string_append(info,"\n");
  }
} 
#endif   
